/**
 * @file      ipaddress.hpp
 * @brief     The main include file for the ipaddress library
 * @author    Vladimir Shaleev
 * @copyright MIT License
 * 
 * This file includes all the necessary components for working with IP addresses 
 * and networks, as well as providing functionality for their manipulation.
 */

#ifndef IPADDRESS_IPADDRESS_HPP
#define IPADDRESS_IPADDRESS_HPP

#include "ipv4-address.hpp"
#include "ipv6-address.hpp"
#include "ipv4-network.hpp"
#include "ipv6-network.hpp"
#include "ip-networks.hpp"
#include "ip-any-address.hpp"
#include "ip-any-network.hpp"

/**
 * @namespace ipaddress
 * 
 * Namespace dedicated to IP addresses and networks operations.
 * 
 * This namespace contains all the classes, functions, and utilities necessary for handling
 * both IPv4 and IPv6 addresses and networks. It provides a comprehensive suite of tools
 * for various operations such as parsing, constructing, and manipulating IP addresses and networks.
 */

#endif
